unit Wavemx;

interface

uses
	SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
	Forms, Dialogs, WaveMix;

type
	EWaveMixer = class(Exception);

	TMixerDevice = class(TObject)
	private
		{ Private declarations }
	protected
		{ Protected declarations }
		FHWaveMix : THandle;
		FChannelsOpen : Boolean;
		FNumChannels : Integer;

		procedure SetNumChannels(Value : Integer);

		procedure OpenChannels;
		procedure CloseChannels;
	public
		{ Public declarations }
		constructor Create; virtual;
		destructor Destroy; override;

		property HWaveMix : THandle read FHWaveMix;
		property NumChannels : Integer read FNumChannels write SetNumChannels;
	end;

	TWaveMixer = class(TComponent)
	private
		{ Private declarations }
	protected
		{ Protected declarations }
		FMixWave : PMixWave;
		FFileName : TFileName;
		FLoaded : Boolean;
		FOpened : Boolean;
		FPlaying : Boolean;
		FOnSoundStopped : TNotifyEvent;
		FChannels : TList;
		FWindowHandle: HWND;
		FEnded : Boolean;

		procedure WndProc(var Msg: TMessage);

		function ParseFileName(const AFileName : string) : string; virtual;
		procedure SoundStopped; virtual;
		procedure SetFileName(Value : TFileName);
		procedure SetLoaded(Value : Boolean);

		procedure PlayFile(Looped : Boolean);
		procedure OpenFile;
		procedure CloseFile;
	public
		{ Public declarations }
		constructor Create(AOwner : TComponent); override;
		destructor Destroy; override;

		procedure Play;
		procedure PlayLooped;
		procedure Stop;

		property Playing : Boolean read FPlaying;
		property Channels : TList read FChannels;
		property Ended : Boolean read FEnded;
	published
		{ Published declarations }
		property FileName : TFileName read FFileName write SetFileName;
		property Loaded : Boolean read FLoaded write SetLoaded default False;
		property OnSoundStopped : TNotifyEvent read FOnSoundStopped write FOnSoundStopped;
	end;

{*****************************************************************
	WaveMix Distribution:
		WaveMix.DLL 		-> Windows/System or Program Directory
		WaveMix.INI			-> Windows or Program Directory
*****************************************************************}

procedure Register;

implementation

uses
	mmSystem, IniFiles;

procedure ShowMMError(Value : Word);
var
	S : String;
	IniFile : TIniFile;
begin
	if Value <> 0 then
	begin
		case Value of
			mmsyserr_Error : S := 'General WaveMixer Error';
			mmsyserr_BadDeviceID : S := 'Device ID out of range';
			mmsyserr_NotEnabled : S := 'Driver failed enable';
			mmsyserr_Allocated : S := 'Device already allocated';
			mmsyserr_InvalHandle : S := 'Device handle is invalid';
			mmsyserr_NoDriver : S := 'No device driver present';
			mmsyserr_NoMem : S := 'Memory allocation error';
			mmsyserr_NotSupported : S := 'Function isn''t supported';
			mmsyserr_BadErrNum : S := 'Error value out of range';
			mmsyserr_InvalFlag : S := 'Invalid Flag Passed';
			mmsyserr_InvalParam : S := 'Invalid Parameter';
			else
				S := 'Unknown WaveMixer Error';
		end;
{		raise EWaveMixer.Create(S);}
{		IniFile := TIniFile.Create('C:\MIXERROR.INI');
		try
			IniFile.WriteString('Errors', S, '1');
		finally
			IniFile.Free;
		end;}
	end;
end;

var
	MixerDevice : TMixerDevice;
	MixerDeviceCount : Word;

procedure OpenMixerDevice;
begin
	if MixerDeviceCount = 0 then
		MixerDevice := TMixerDevice.Create;

	Inc(MixerDeviceCount);
end;

procedure CloseMixerDevice;
begin
	Dec(MixerDeviceCount);

	if MixerDeviceCount = 0 then
		MixerDevice.Free;
end;


{ TMixerDevice }

constructor TMixerDevice.Create;
begin
	FHWaveMix := WaveMixInit;
	if FHWaveMix = 0 then
		raise EWaveMixer.Create('WaveMix not created');
	FNumChannels := -1;
	FChannelsOpen := False;
	OpenChannels;
	ShowMMError(WaveMixActivate(FHWaveMix, True))
end;

destructor TMixerDevice.Destroy;
begin
	CloseChannels;
	WaveMixCloseSession(FHWaveMix);
	inherited Destroy;
end;

procedure TMixerDevice.OpenChannels;
var
	Flags : Word;
begin
	if not FChannelsOpen then
	begin
		if FNumChannels < 0 then
			Flags := WMIX_OPENALL
		else
			Flags := WMIX_OPENCOUNT;
		ShowMMError(WaveMixOpenChannel(FHWaveMix, FNumChannels, Flags));
		FChannelsOpen := True;
	end;
end;

procedure TMixerDevice.CloseChannels;
begin
	if FChannelsOpen then
	begin
		ShowMMError(WaveMixCloseChannel(FHWaveMix, 0, WMIX_ALL));
		FChannelsOpen := False;
	end;
end;

procedure TMixerDevice.SetNumChannels(Value : Integer);
begin
	if Value <> FNumChannels then
	begin
		CloseChannels;
		FNumChannels := Value;
		OpenChannels;
	end;
end;

{ TWaveMixer }

constructor TWaveMixer.Create(AOwner : TComponent);
begin
	inherited Create(AOwner);
	FMixWave := nil;
	FOpened := False;
	FLoaded := False;
	FPlaying := False;
	FEnded := True;
	FChannels := TList.Create;
	FWindowHandle := AllocateHWnd(WndProc);
end;

destructor TWaveMixer.Destroy;
begin
	CloseFile;
	FChannels.Free;
	DeallocateHWnd(FWindowHandle);
	inherited Destroy;
end;

procedure TWaveMixer.WndProc(var Msg: TMessage);
var
	ChannelIdx : Integer;
begin
	with Msg do
		if Msg = MM_WOM_DONE then
			try
				ChannelIdx := FChannels.IndexOf(Pointer(wParam));
				if (ChannelIdx > -1) and (Pointer(lParam) = FMixWave) then
				begin
					FChannels.Delete(ChannelIdx);
					FEnded := True;
					SoundStopped;
				end;
{				MessageDlg(Format('wParam = %d, lParam = %p, Channel = %d, Handle = %d, Pointer = %p',
						[wParam, Pointer(lParam), FChannel, MixerDevice.HWaveMix, FMixWave]), mtInformation, [mbOK], 0);}
			except
				Application.HandleException(Self);
			end
		else
			Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

function TWaveMixer.ParseFileName(const AFileName : string) : string;
var
	n : integer;
begin
	Result := AFileName;
	n := Pos('%P', UpperCase(AFileName));
	if n > 0 then
	begin
		Delete(Result, n, 2);
		while Result[n] = '\' do
			Delete(Result, n, 1);
		Insert(ExtractFilePath(Application.ExeName), Result, n)
	end;
end;

procedure TWaveMixer.SoundStopped;
begin
	if FChannels.Count = 0 then
		FPlaying := False;
	if not FLoaded then
		CloseFile;

	if Assigned(FOnSoundStopped) then
		FOnSoundStopped(Self);
end;

procedure TWaveMixer.SetFileName(Value : TFileName);
begin
	if Value <> FFileName then
	begin
		CloseFile;
		FFileName := Value;
		if FLoaded then
			OpenFile;
	end;
end;

procedure TWaveMixer.SetLoaded(Value : Boolean);
begin
	if Value <> FLoaded then
	begin
		if Value then
			if not (csDesigning in ComponentState) then
				OpenFile
			else
		else
			CloseFile;
		FLoaded := Value;
	end;
end;

procedure TWaveMixer.OpenFile;
var
	NullFileName : array[0..150] of char;
	FN : string;
begin
	if not FOpened and not (csDesigning in ComponentState) then
	begin
		if FFileName = '' then
			EWaveMixer.Create('No Filename specified');
		FN := ParseFileName(FFileName);
		if not FileExists(FN) then
			raise EWaveMixer.CreateFmt('Unable to open wave file: %s', [FN]);

		OpenMixerDevice;

		StrPCopy(NullFileName, FN);
		FMixWave := WaveMixOpenWave(MixerDevice.HWaveMix, @NullFileName, 0, WMIX_FILE);
		if FMixWave = nil then
			raise EWaveMixer.CreateFmt('Unable to open wave file: %s', [FileName]);

		FOpened := True;
	end;
end;

procedure TWaveMixer.CloseFile;
begin
	if FOpened then
	begin
		ShowMMError(WaveMixFreeWave(MixerDevice.HWaveMix, FMixWave));

		CloseMixerDevice;

		FOpened := False;
	end;
end;

procedure TWaveMixer.PlayFile(Looped : Boolean);
var
	MixPlayParams : TMixPlayParams;
begin
	OpenFile;

	with MixPlayParams do
	begin
		wSize := Sizeof(MixPlayParams);
		hMixSession := MixerDevice.HWaveMix;
		iChannel := 0;
		lpMixWave := FMixWave;
		hWndNotify := FWindowHandle;
		dwFlags := WMIX_USELRUCHANNEL or WMIX_QUEUEWAVE{WMIX_CLEARQUEUE};
		if Looped then
			wLoops := $FFFF
		else
			wLoops := 0;
	end;
	ShowMMError(WaveMixPlay(@MixPlayParams));
	FChannels.Add(Pointer(MixPlayParams.iChannel));
	FPlaying := True;
end;

procedure TWaveMixer.Play;
begin
	PlayFile(False);
end;

procedure TWaveMixer.PlayLooped;
begin
	PlayFile(True);
end;

procedure TWaveMixer.Stop;
var
	i : integer;
begin
	for i := 0 to FChannels.Count - 1 do
		ShowMMError(WaveMixFlushChannel(MixerDevice.HWaveMix, Integer(FChannels[i]), 0));
	if FChannels.Count > 0 then
	begin
		FChannels.Clear;
		FEnded := False;
		SoundStopped;
	end;
	FPlaying := False;
end;

procedure Register;
begin
	RegisterComponents('Tigh Tech', [TWaveMixer]);
end;

begin
	MixerDevice := nil;
	MixerDeviceCount := 0;
end.
